% VAR estimation and simulation of residuals. Includes back-transform of
% residuals

clear all

% General parameters
start_year = 1990; % Start year of time series
end_year = 2015; % Final year of time series, including this year
year_length = 2920; % 1460 Number of observations in a year
filename = 'WAM10.txt'; % WAM 10 Metocean file without headers. 
harsh = 1; % 0 for calm, 1 for 
num_rep = 10; % Number of replications 
write_to_csv = 0; % write residuals
write_filename = 'VAR_XX_MS_C.csv';

% VAR paramaters
np = 10; % Number of p lags to be evaluated. Integer and larger than zero
burnIn = 400; % Number of time steps for burn-in. 200 is normal

% Load Hs, Tp, year indices and month indices from start year to and 
% including end year
if harsh == 1
    metocean = loadTimeSeries(start_year,end_year,filename);
else
    metocean = csvread(filename);
end

% Length of time series
series_length = length(metocean(:,1));

%% Uniformly distribute Tp
if harsh == 1
    % Spread Tp with uniform distribution
    Tp_spread = zeros(length(metocean),1);
    for ii = 1:length(metocean(:,4))
        n = round(1+log(metocean(ii,4)/3.244)/0.09525);
        metocean(ii,4) = 3.244*exp(0.09525*(n-0.5-rand));
    end
else 
    metocean(:,4) = spread_calm_peak_period(metocean);
end

%% Transformation to normality
% Allocate year and month indices
metocean_norm(:,1) = metocean(:,1);
metocean_norm(:,2) = metocean(:,2);

% Rosenblatt transform
%Creating scatter diagram of 
scatter_diagram = scatterDiagram(metocean(:,3),metocean(:,4));

% Determine regression coefficients for mu and sigma for lognormal 
[a, b] = meanAndStdDev(scatter_diagram,0);

% Fit lognormal distribution to Hs and find parameters mu and sigma
logn_par = lognfit(metocean(:,3));

% Transform Hs to lognormal probability
F_Hs = logncdf(metocean(:,3),logn_par(1),logn_par(2));

% Transform Tp to lognormal probability conditioned on Hs
F_Tp = zeros(length(metocean(:,3)),1);
for ii = 1:length(metocean(:,3))
    % Conditioned mu and sigma values 
    mu = a(1) + a(2)*(metocean(ii,3)^a(3));
    sigma = sqrt(b(1) + b(2)*exp(-b(3)*metocean(ii,3)));
    F_Tp(ii) = logncdf(metocean(ii,4),mu,sigma);
end

% Inverse standard normal trasnform of Hs and Tp
[metocean_norm(:,3)] = norminv(F_Hs);
[metocean_norm(:,4)] = norminv(F_Tp);

%% Seasonal Transformation
[yearMeanFit,yearCovFit,midMonthHours] = seasonal_trans(start_year,end_year,metocean_norm,year_length);

% Fit Fourier series to monthly mean and covariance
M_Hs = fit(midMonthHours,yearMeanFit(:,1),'fourier6'); % M Hs
M_Tp = fit(midMonthHours,yearMeanFit(:,2),'fourier6'); % M Tp
S_11 = fit(midMonthHours,yearCovFit(:,1),'fourier6');
S_12 = fit(midMonthHours,yearCovFit(:,2),'fourier6'); % S12=S21
S_22 = fit(midMonthHours,yearCovFit(:,4),'fourier6');

%% Generate Residuals
% Pre-allocate
W = zeros(series_length,2);

% Perform seasonal transformation W = (Y-M)/S
for ii = 1:series_length
    Y = [metocean_norm(ii,3); metocean_norm(ii,4)];
    M = [M_Hs(ii); M_Tp(ii)];
    S = sqrt([S_11(ii) 0; 0 S_22(ii)]);
    W(ii,:)=(Y-M);
    %W(ii,:)=S\(Y-M); % x = A\b --- Ax = b
end

%% Find optimal VAR model lags for residuals W with AIC
% Initilize
LOGL = zeros(np,1);

% Fit VAR(p) models
for p = 1:np
        mdl = varm(2,p);
        [fit,~,logL] = estimate(mdl,[W(:,1) W(:,2)]);
        % Store the loglikelihood objective function and number of 
        % coefficients for each fitted model
        LOGL(p) = logL;
end

% Column vector with number of terms for each VAR(p)
P(:,1) = 2+4*[1:np]; 

% Calculate the AIC/BIC for each fitted VAR model
[AIC,BIC] = aicbic(LOGL,P,series_length);

% Optimal number of lags according to AIC. 
[minNumAIC,pLagsAIC] = min(AIC);

%% Simulate Residuals W
% Re-fit optimal VAR model to W residuals
mdlW = varm(2,pLagsAIC);
estMdlW = estimate(mdlW,[W(:,1) W(:,2)]);

% Simulate series
WsimBurnIn = simulate(estMdlW,burnIn+series_length,'NumPaths',num_rep);
W_sim = WsimBurnIn(burnIn+1:end,:,:);

%% Inverse Seasonal Transform Y = M + S*W
Y_sim_norm = zeros(series_length,2,num_rep);
for ii = 1:num_rep
    for jj = 1:series_length
        M = [M_Hs(jj); M_Tp(jj)];
        S = sqrt([S_11(ii) 0; 0 S_22(ii)]);
        W_inv = [W_sim(jj,1,ii); W_sim(jj,2,ii)];
        %Y_sim_norm(jj,:,ii) = W_inv + M;
        Y_sim_norm(jj,:,ii) = M + S*W_inv;
    end
end

%% Inverse Rosenblatt transform
% Standard normal transform
Y_sim_prob = zeros(series_length,2,num_rep);
for ii = 1:num_rep
    Y_sim_prob(:,1:2,ii) = normcdf(Y_sim_norm(:,1:2,ii));
end

% Inverse lognormal CDF transform
Y_sim = zeros(series_length,2,num_rep);
for ii = 1:num_rep
    Y_sim(:,1,ii) = logninv(Y_sim_prob(:,1,ii),logn_par(1),logn_par(2));
    for jj = 1:series_length
    % mu and sigma vectors for Tp Lognormal transform
    mu = a(1) + a(2)*(Y_sim(jj,1,ii)^a(3));
    sigma = sqrt(b(1) + b(2)*exp(-b(3)*Y_sim(jj,1,ii)));
    Y_sim(jj,2,ii) = logninv(Y_sim_prob(jj,2,ii),mu,sigma);
    end
end

%% Write simulated series to .csv
if write_to_csv == 1
    indices = [metocean(:,1) metocean(:,2)];
    for ii = 1:num_rep
        if ii == 1
            y_sim_out = [Y_sim(:,1,ii) Y_sim(:,2,ii)];
        else
            y_sim_out = [y_sim_out Y_sim(:,1,ii) Y_sim(:,2,ii)];
        end
    end
    out = [indices y_sim_out];
    csvwrite(write_filename,out)
end
