% Generate Residuals
clear all
clc

% Parameters
start_year = 1988; % Start year of time series
end_year = 2015; % Final year of time series, including this year
filename = 'WAM10.txt'; % WAM 10 Metocean file without headers. 
harsh = 1; % Use any value for calm. 
rosen = 1; % 

% Load Hs, Tp, year indices and month indices from start year to and 
% including end year
if harsh == 1
    metocean = loadTimeSeries(start_year,end_year,filename);
    year_length = 2920;
else
    metocean = csvread(filename);
    year_length = 1460;
end

% Length of time series
series_length = length(metocean(:,1));


%% Uniformly distribute Tp
if harsh == 1
    % Spread Tp with uniform distribution
    Tp_spread = zeros(length(metocean),1);
    for ii = 1:length(metocean(:,4))
        n = round(1+log(metocean(ii,4)/3.244)/0.09525);
        metocean(ii,4) = 3.244*exp(0.09525*(n-0.5-rand));
    end
else 
    metocean(:,4) = spread_calm_peak_period(metocean);
end

%% Transformation to normality
% Allocate year and month indices
metocean_norm(:,1) = metocean(:,1);
metocean_norm(:,2) = metocean(:,2);

if rosen == 1
    % Rosenblatt transform
    %Creating scatter diagram of 
    scatter_diagram = scatterDiagram(metocean(:,3),metocean(:,4));

    % Determine regression coefficients for mu and sigma for lognormal 
    [a, b] = meanAndStdDev(scatter_diagram,0);

    % Fit lognormal distribution to Hs and find parameters mu and sigma
    logn_par = lognfit(metocean(:,3));

    % Transform Hs to lognormal probability
    F_Hs = logncdf(metocean(:,3),logn_par(1),logn_par(2));

    % Transform Tp to lognormal probability conditioned on Hs
    F_Tp = zeros(length(metocean(:,3)),1);
    for ii = 1:length(metocean(:,3))
        % Conditioned mu and sigma values 
        mu = a(1) + a(2)*(metocean(ii,3)^a(3));
        sigma = sqrt(b(1) + b(2)*exp(-b(3)*metocean(ii,3)));
        F_Tp(ii) = logncdf(metocean(ii,4),mu,sigma);
    end
    % Inverse standard normal trasnform of Hs and Tp
    [metocean_norm(:,3)] = norminv(F_Hs);
    [metocean_norm(:,4)] = norminv(F_Tp);
else
    % Multivariate BoxCox transform
    lambda_cox = boxcoxn(metocean(:,3:4));
    [metocean_norm(:,3)] = boxcox(lambda_cox(1),metocean(:,3));
    [metocean_norm(:,4)]= boxcox(lambda_cox(2),metocean(:,4));
end

%% Seasonal Transformation
[yearMeanFit,yearCovFit,midMonthHours] = seasonal_trans(start_year,end_year,metocean_norm,year_length);

% Fit Fourier series to monthly mean and covariance
M_Hs = fit(midMonthHours,yearMeanFit(:,1),'fourier6'); % M Hs
M_Tp = fit(midMonthHours,yearMeanFit(:,2),'fourier6'); % M Tp
S_11 = fit(midMonthHours,yearCovFit(:,1),'fourier6');
S_12 = fit(midMonthHours,yearCovFit(:,2),'fourier6'); % S12=S21
S_22 = fit(midMonthHours,yearCovFit(:,4),'fourier6');

%% Generate Residuals
% Pre-allocate
W = zeros(series_length,2);

% Perform seasonal transformation W = (Y-M)/S
for ii = 1:series_length
    Y = [metocean_norm(ii,3); metocean_norm(ii,4)];
    M = [M_Hs(ii); M_Tp(ii)];
    S = sqrt([S_11(ii) 0; 0 S_22(ii)]);
    %W(ii,:)=(Y-M);
    W(ii,:)=S\(Y-M); % x = A\b --- Ax = b
end

% Write residuals W to .csv file
out = [metocean(:,1:2) W];

%out = out(out(:,1) == 1989,:);
writeFileName = 'residuals_pred_test.csv';
csvwrite(writeFileName,out)

%% Validation Plots
% Plot of Fourier approximations for validation of fit
figure(1)
subplot(1,2,1)
plot(M_Hs,midMonthHours,yearMeanFit(:,1))
title('M Hs')
subplot(1,2,2)
plot(M_Tp,midMonthHours,yearMeanFit(:,2))
title('M Tp')

figure(2)
plot(S_11,midMonthHours,yearCovFit(:,1))
hold on
plot(S_12,midMonthHours,yearCovFit(:,2))
hold on
plot(S_22,midMonthHours,yearCovFit(:,4))
hold off

if rosen == 1
    figure(3)
    subplot(1,2,1)
    histogram(norminv(F_Hs))
    subplot(1,2,2)
    histogram(norminv(F_Tp))

    figure(4)
    scatter(norminv(F_Hs),norminv(F_Tp))
    xlabel('u1')
    ylabel('u2')
    title('Residuals')
end
