% Markov model

clear all

startYear = 1990; % Start year of time series
endYear = 2015; % Final year of time series, including this year
filename = 'WAM10.txt'; % WAM 10 Metocean file without headers. 
numStates = 24; % Set number of states for Hs
numRep = 10;
year_length = 2920;
harsh = 1;

% Load series
if harsh == 1
    metocean = loadTimeSeries(startYear,endYear,filename);
else
    metocean = csvread(filename);
end

% Load Hs and Tp for the last 30 years from hindcast data
Hs = round(metocean(:,3),1);
Tp = metocean(:,4);

% Length of time series
seriesLength = length(metocean(:,1));
% Number of Replications
numReplications = seriesLength; 

% Generate the coupling matrix
C = CouplingMatrix(Hs,Tp); 

%% Seasonal Transformation

% Initilize
meanHsMonth = zeros(1,12); stdHsMonth = zeros(1,12);

% Create vector with monthly mean and variance from observed time series
for ii = 1:12
meanHsMonth(ii) = mean(Hs(metocean(:,2) == ii));
stdHsMonth(ii) = std(Hs(metocean(:,2) == ii));
end

% Vector containing the middle day of each month in hours from year start
% Day 1 = 365, such that the period is equal to one year 
midMonthHours = transpose([1 15.5 45 74.5 105 135.5 166 196.5 ...
    227.5 258 288.5 319 349.5 365]*(year_length/365));

% Row containing the average mean value for the 1st and 365th day
% Found as the mean of month 1 and month 12
startEndMean = (meanHsMonth(1) + meanHsMonth(end))/2;
startEndSTD = (stdHsMonth(1) + stdHsMonth(end))/2;

% Add the average value row as the first and last element
MeanFit = transpose([startEndMean  meanHsMonth startEndMean]);
stdFit = transpose([startEndMean  stdHsMonth startEndMean]);

% Fit Fourier series to monthly mean values i.e., find periodic M(t)
M = fit(midMonthHours,MeanFit,'fourier6'); % M Hs
S = fit(midMonthHours,stdFit,'fourier5'); % S Hs

% Plot of Fourier approximations for validation of fit
%{
figure(1)
subplot(1,2,1)
plot(M,midMonthHours,MeanFit)
title('M Hs')
subplot(1,2,2)
plot(S,midMonthHours,stdFit)
title('S Hs')
%}
%% Seasonal transformation

% Initilize
Wobs = zeros(seriesLength,1);

% Perform seasonal transformation W = (Y-M)/S
for ii = 1:seriesLength
    Wobs(ii) = (Hs(ii)-M(ii));%/S(ii);
end

%% ------------------- P Matrix ----------------------------------
% Find upper and lower limits for Hs values
ub = max(Wobs);
lb = min(Wobs);

% Find state ranges
stateRange = (ub-lb) / (numStates);
% State values
stateValues = lb+stateRange:stateRange:ub;
% Preallocation
State = zeros(seriesLength,1);

% Find each data points state
for i = 1:length(Wobs)
    % For each data point
    for j = 1:numStates
        % For each state
        if Wobs(i) <= stateValues(j)
            % Data point is in state j
            State(i) = j;
            % This data point is categorized, so we break and move to the
            % next data point
            break;
        end
    end
end

% Find transitions
transitions = zeros(numStates);

for t = 1:seriesLength-1
    % HsState(t) represents the state and HsState(t+1) represents the state
    % it transitions to
    transitions(State(t),State(t+1)) = transitions(State(t),State(t+1)) + 1;
end

P = transitions;
% Normalize each row in the transition matrix so each row sums to 1
for i = 1:numStates
    P(i,:) = P(i,:) / sum( P(i,:) );
end

% Check to see if there are any absorbing states
% i.e. P(i,j) == 1 where i=j
absorbstate = zeros(numStates);
for i = 1:numStates
    for j = 1:numStates
        if P(i,j) == 1
            absorbstate(i,j) = absorbstate(i,j) + 1;
        end
    end
end
if sum(sum(absorbstate)) >= 1
    error('Absorbing states');
end

%% Simulation
% Randomly sample start state
state = randi(numStates);

% Preallocation
states = zeros(seriesLength,1);

for kk = 1:numRep
    for i = 1:seriesLength
        % Sample a new random value in range [0,1]
        r = rand();

        for j = 1:numStates
            prob = 0;
            % Accumulate probabilities
            for k = 1:j
                prob = prob + P(state,k);
            end

            if r <= prob
                % New state is found, j
                state = j;

                % Store the state we transition to
                states(i,kk) = j;

                % Break ends the current for loop, and returns to the outer
                % loop, which will sample a new random value and start over
                break;
            end
        end
    end
end

% Transform states to normalized values
WSim = zeros(seriesLength,numRep);
for jj = 1:numRep
    for ii = 1:seriesLength
        WSim(ii,jj) = stateValues(states(ii,jj));
    end
end

%% -------------------------- Inverse Transform ---------------------------
% Preallocation
HsSim = zeros(seriesLength,numRep);

% Apply inverse 
for jj = 1:numRep
    for ii = 1:seriesLength
        HsSim(ii,jj) = M(ii) + WSim(ii,jj);%*S(ii);
        if HsSim(ii,jj) < min(Hs)
            HsSim(ii,jj) = min(Hs);
        end
    end
end

%% -------------------------- Coupling ----------------------------------
Tp_unique = unique(Tp); Hs_unique = unique(Hs);
n_unique_Hs = length(Hs_unique); n_unique_Tp = length(Tp_unique);
simTpSpread = zeros(seriesLength,numRep);

for kk = 1:numRep
% Preallocation
Hs_state = zeros(seriesLength,1);

% Find closest unique Hs state
for ii = 1:seriesLength 
    for jj = 1: n_unique_Hs 
        if HsSim(ii,kk) <= Hs_unique(jj)
            Hs_state(ii) = jj;
            % Break inner loop when state is found
            break;
        end
    end
end

% Can try to enter state max + 1
Hs_state(Hs_state == 0) = max(Hs_state);

% Preallocate Tp coupled vector
Tp_coupled = zeros(1,seriesLength);

for jj = 1:seriesLength
    
r = rand; % Draw random number between 0 and 1
prob = 0; % Reset probabilty for each replication

    for ii = 1:n_unique_Tp
        % Accumulate probabilities 
        prob = prob + C(Hs_state(jj),ii);
        
        % Find Tp values
        if r <= prob
            Tp_coupled(jj) = Tp_unique(ii);
            % Break inner loop when state is found
            break;
        end
    end
end

%% Spread T_p Values

if harsh == 1
    % Simulated values 
    % Preallocation
    Tp_coupled_spread = zeros(1,length(Tp_coupled));

    % Apply spreading function to simulated Tp
    for ii = 1:length(Tp_coupled)
        n = round(1+log(Tp_coupled(ii)/3.244)/0.09525);
        Tp_coupled_spread(ii) = 3.244*exp(0.09525*(n-0.5-rand));
    end

    simTpSpread(:,kk) = Tp_coupled_spread;
    
else
    metocean(:,4) = Tp_coupled;
    simTpSpread(:,kk) = spread_calm_peak_period(metocean);
end
end    

if harsh == 1
    % Observed values spread_calm_peak_period(metocean)
    % Preallocation
    Tp_spread = zeros(length(Tp),1);

    % Apply spreading function to observed Tp
    for ii = 1:length(Tp)
        n = round(1+log(Tp(ii)/3.244)/0.09525);
        Tp_spread(ii,1) = 3.244*exp(0.09525*(n-0.5-rand));
    end
else
    metocean(:,4) = Tp;
    Tp_spread = spread_calm_peak_period(metocean);
end

%% Write Series Ysim to csv

Ysim = zeros(seriesLength,2,numRep);

for ii = 1:numRep
Ysim(:,1,ii) = HsSim(:,ii);
Ysim(:,2,ii) = simTpSpread(:,ii);
end

indices = [metocean(:,1) metocean(:,2)];
for ii = 1:numRep
    if ii == 1
        sims = [Ysim(:,1,ii) Ysim(:,2,ii)];
    else
        sims = [sims Ysim(:,1,ii) Ysim(:,2,ii)];
    end
end
out = [indices sims];
csvwrite('Markov_M_H.csv',out)

