% Back-transform residuals

clear all

% Parameters
start_year = 1990; % Start year of time series
end_year = 2015; % Final year of time series, including this year
observed_filename = 'WAM10.txt'; % WAM 10 Metocean file without headers. 
residual_filename = 'Residuals_VARMA_31_BOXCOX_MS_H.csv';
num_rep = 10;
harsh = 1;
std_on = 1;
write = 0;
rosen = 0;

% Load Hs, Tp, year indices and month indices from start year to and 
% including end year
if harsh == 1
    metocean = loadTimeSeries(start_year,end_year,observed_filename);
    year_length = 2920;
else
    metocean = csvread(observed_filename);
    year_length = 1460;
end

% Load residual series
W_sim_raw = load(residual_filename);

% Length of time series
series_length = length(W_sim_raw);

% Allocate year and month indices
metocean_norm(:,1) = metocean(:,1);
metocean_norm(:,2) = metocean(:,2);

%% Uniformly distribute Tp
if harsh == 1
    % Spread Tp with uniform distribution
    Tp_spread = zeros(length(metocean),1);
    for ii = 1:length(metocean(:,4))
        n = round(1+log(metocean(ii,4)/3.244)/0.09525);
        metocean(ii,4) = 3.244*exp(0.09525*(n-0.5-rand));
    end
else 
    metocean(:,4) = spread_calm_peak_period(metocean);
end

%% Transformation to normality
% Allocate year and month indices
metocean_norm(:,1) = metocean(:,1);
metocean_norm(:,2) = metocean(:,2);

if rosen == 1
    % Rosenblatt transform
    %Creating scatter diagram of 
    scatter_diagram = scatterDiagram(metocean(:,3),metocean(:,4));

    % Determine regression coefficients for mu and sigma for lognormal 
    [a, b] = meanAndStdDev(scatter_diagram,0);

    % Fit lognormal distribution to Hs and find parameters mu and sigma
    logn_par = lognfit(metocean(:,3));

    % Transform Hs to lognormal probability
    F_Hs = logncdf(metocean(:,3),logn_par(1),logn_par(2));

    % Transform Tp to lognormal probability conditioned on Hs
    F_Tp = zeros(length(metocean(:,3)),1);
    for ii = 1:length(metocean(:,3))
        % Conditioned mu and sigma values 
        mu = a(1) + a(2)*(metocean(ii,3)^a(3));
        sigma = sqrt(b(1) + b(2)*exp(-b(3)*metocean(ii,3)));
        F_Tp(ii) = logncdf(metocean(ii,4),mu,sigma);
    end
    % Inverse standard normal trasnform of Hs and Tp
    [metocean_norm(:,3)] = norminv(F_Hs);
    [metocean_norm(:,4)] = norminv(F_Tp);
else
    % Multivariate BoxCox transform
    lambda_cox = boxcoxn(metocean(:,3:4));
    [metocean_norm(:,3)] = boxcox(lambda_cox(1),metocean(:,3));
    [metocean_norm(:,4)]= boxcox(lambda_cox(2),metocean(:,4));
end

%% Inverse Seasonal Transformation
[yearMeanFit,yearCovFit,midMonthHours] = seasonal_trans(start_year,end_year,metocean_norm,year_length);

% Fit Fourier series to monthly mean and covariance
M_Hs = fit(midMonthHours,yearMeanFit(:,1),'fourier6'); % M Hs
M_Tp = fit(midMonthHours,yearMeanFit(:,2),'fourier6'); % M Tp
S_11 = fit(midMonthHours,yearCovFit(:,1),'fourier6');
S_12 = fit(midMonthHours,yearCovFit(:,2),'fourier6'); % S12=S21
S_22 = fit(midMonthHours,yearCovFit(:,4),'fourier6');

% Reshape simulated values
W_sim = zeros(series_length,2,num_rep);

for ii = 1:num_rep
    W_sim(:,1,ii) = W_sim_raw(:,2*(ii-1)+1);
    W_sim(:,2,ii) = W_sim_raw(:,2*ii);
end

% Inverse seasonal transform
Y_sim_norm = zeros(series_length,2,num_rep);
if std_on == 1
    for ii = 1:num_rep
        for jj = 1:series_length
            M = [M_Hs(jj); M_Tp(jj)];
            S = sqrt([S_11(ii) 0; 0 S_22(ii)]);
            W_inv = [W_sim(jj,1,ii); W_sim(jj,2,ii)];
            Y_sim_norm(jj,:,ii) = M + S*W_inv;
        end
    end
else
    for ii = 1:num_rep
        for jj = 1:series_length
            M = [M_Hs(jj); M_Tp(jj)];
            W_inv = [W_sim(jj,1,ii); W_sim(jj,2,ii)];
            Y_sim_norm(jj,:,ii) = W_inv + M;
        end
    end
end

%% Inverse Rosenblatt transform
Y_sim = zeros(series_length,2,num_rep);
    
% Standard normal transform
if rosen == 1
    Y_sim_prob = zeros(series_length,2,num_rep);
    % Standard Normal CDF
    for ii = 1:num_rep
        Y_sim_prob(:,1:2,ii) = normcdf(Y_sim_norm(:,1:2,ii));
    end

    % Inverse lognormal CDF transform

    for ii = 1:num_rep
        Y_sim(:,1,ii) = logninv(Y_sim_prob(:,1,ii),logn_par(1),logn_par(2));
        for jj = 1:series_length
            % mu and sigma vectors for Tp Lognormal transform
            mu = a(1) + a(2)*(Y_sim(jj,1,ii)^a(3));
            sigma = sqrt(b(1) + b(2)*exp(-b(3)*Y_sim(jj,1,ii)));
            Y_sim(jj,2,ii) = logninv(Y_sim_prob(jj,2,ii),mu,sigma);
            if Y_sim(jj,2,ii) == inf
               Y_sim(jj,2,ii) = Y_sim(jj,2,ii-1);
            end
       end
    end
else
    for ii = 1:num_rep
        Y_sim(:,1,ii) = (lambda_cox(1).*Y_sim_norm(:,1,ii) + 1).^(1/lambda_cox(1));
        Y_sim(:,2,ii) = (lambda_cox(2).*Y_sim_norm(:,2,ii) + 1).^(1/lambda_cox(2));
    end
end


%% Write results to csv
if write == 1
    Y_sim_out = zeros(length(Y_sim),2*num_rep);
    for ii = 1:num_rep
        Y_sim_out(:,2*(ii-1)+1) = Y_sim(:,1,ii);
        Y_sim_out(:,2*ii) = Y_sim(:,2,ii);
    end

    Y_sim_out = [metocean(:,1:2),Y_sim_out];

    csvwrite('VARMA_13_BOXCOX_MS_H.csv',Y_sim_out)
end

beep