% Validate a single synthetic series against observed

clear all

% General parameters
harsh = 1; % 0 for calm, 1 for harsh
num_rep = 10;
lstm = 0;

% plot parameters
m_type = 'VARMA(3,3)$_m$';
line_width = 1.5;
num_lags_auto = 100; % Number of lags in auto/crosscorrelation
plot_pers = 1;

% Files
 % WAM 10 Metocean file without headers. 
synthetic_filename = 'VARMA_33_M_H.csv';

% LOAD PERSISTENCE TABLES
if plot_pers == 1
    obs_persistence = xlsread('pers_table_harsh.xlsx');
    obs_above = obs_persistence(2:61,1:21);
    obs_below = obs_persistence(63:122,1:21);
    syn_persistence = xlsread('pers_table_VARMA_33_M_H.xlsx');
    syn_above = syn_persistence(2:61,1:21);
    syn_below = syn_persistence(63:122,1:21);
end

% Read observed series
if harsh == 1
    observed_filename = 'WAM10.txt';
    step_length = 3;
    start_year = 1990; % Start year of time series
    end_year = 2015; % Final year of time series, including this year
    year_length = 2920; % 1460 Number of observations in a year
    metocean = loadTimeSeries(start_year,end_year,observed_filename);
    scatter_max_x = 22;
    above_values = [.001 .01 .1 .2 .4 .6 .7]; % [.001 .01 .1 .2 .4 .6 .7];
    below_values = [.01 .1 .2 .4 .6 .7 .8]; %[.01 .1 .2 .4 .6 .7 .8];
else
    observed_filename = 'calm_weather.csv';
    metocean = csvread(observed_filename);
    step_length = 6;
    year_length = 1460;
    scatter_max_x = 16;
    above_values = [.001 .01 .05 .1 .2 .6 .7];
    below_values = [.15 .3 .45 .65 .8 .9 .9];
end

%% Set Latex as interpreter and font size
set(0,'defaultTextInterpreter','latex')
set(0, 'defaultAxesTickLabelInterpreter','latex'); 
set(0, 'defaultLegendInterpreter','latex');

set(0,'DefaultAxesFontSize', 16) % 12 for 0.8, 16 for 0.65
set(0,'DefaultTextFontSize', 16)
set(0,'DefaultLegendFontSize', 16)

%% Read 
% Read synthetic series
syn_series = csvread(synthetic_filename);

% Remove month and year columns
syn_series = syn_series(:,3:end);

% Reshape synthetic series
Y_sim = zeros(length(syn_series),2,num_rep);
for ii = 1:num_rep
    Y_sim(:,1,ii) = syn_series(:,2*(ii-1)+1);
    Y_sim(:,2,ii) = syn_series(:,2*ii);
end


%% Uniformly distribute Tp
if harsh == 1
    % Spread Tp with uniform distribution
    for ii = 1:length(metocean(:,4))
        n = round(1+log(metocean(ii,4)/3.244)/0.09525);
        metocean(ii,4) = 3.244*exp(0.09525*(n-0.5-rand));
    end
else 
    metocean(:,4) = spread_calm_peak_period(metocean);
end

Hs_obs = metocean(:,3);
Tp_obs = metocean(:,4);

%% Overall and Monthly Mean and Covaraince
% Over-all mean and covariance
overall_mean_cov = over_all_mean_variance(metocean(:,3),metocean(:,4),Y_sim,num_rep);

% Monthly Mean and Covariance
[obs_cov,sim_cov,obs_month_mean,sim_month_mean] = monthly_mean_covariance(metocean,num_rep,Y_sim);

% Plot monthly mean and covariance
figure(1)
clf
plot(1:12,obs_month_mean(1,:),'LineWidth',line_width)
hold on
plot(1:12,sim_month_mean(1,:),'LineWidth',line_width,'LineStyle','--')
plot(1:12,obs_month_mean(2,:),'LineWidth',line_width)
plot(1:12,sim_month_mean(2,:),'LineWidth',line_width,'LineStyle','--')
legend({'Observed $H_s$',[m_type ' $H_s$'],'Observed $T_p$',[m_type ' $T_p$']})
% title([m_type ' Comparison Plot of Monthly Mean for Series ' s_type])
xlabel('Month','Color','k')
ylabel('Mean','Color','k')
xticks([1 3 6 9 12])
xlim([1,12])
if harsh == 1
    ylim([0,14])
else
    ylim([0,10.5]) % 14 NOT LSTM
end
grid on


figure(2)
clf
plot(1:12,obs_cov(1,:),'LineWidth',line_width)
hold on
plot(1:12,sim_cov(1,:),'LineWidth',line_width,'LineStyle','--')
plot(1:12,obs_cov(2,:),'LineWidth',line_width)
plot(1:12,sim_cov(2,:),'LineWidth',line_width,'LineStyle','--')
plot(1:12,obs_cov(3,:),'LineWidth',line_width)
plot(1:12,sim_cov(3,:),'LineWidth',line_width,'LineStyle','--')
legend('Observed $\Sigma_{11}$',[m_type ' $\Sigma_{11}$'],'Observed $\Sigma_{12}$',[m_type ' $\Sigma_{12}$'],'Observed $\Sigma_{22}$',[m_type ' $\Sigma_{22}$']);
%title('Comparison Plot of Monthly Covariance')
xlabel('Month','Color','k')
ylabel('Covariance','Color','k')
xticks([1 3 6 9 12])
xlim([1,12])
if harsh == 1
    ylim([0,8]) % 24 LSTM
else
    ylim([0,5.5]) %  30 NOT LSTM
end
grid on

%% Autocorrelation and Crosscorrelation

% Hs Autocorrelation
[acf_Hs_sim,lags_auto] = autocorr(Y_sim(:,1,1),num_lags_auto);
acf_Hs_obs = autocorr(metocean(:,3),num_lags_auto);

% Tp Autocorrelation
acf_Tp_sim = autocorr(Y_sim(:,2,1),num_lags_auto);
acf_Tp_obs = autocorr(metocean(:,4),num_lags_auto);

figure(3)
clf
subplot(2,1,1)
plot(lags_auto,acf_Hs_obs,'LineWidth',line_width)
hold on
plot(lags_auto,acf_Hs_sim,'LineWidth',line_width,'LineStyle','--')
%title('$H_s$ Autocorrelation')
xlabel('Lag','Color','k')
ylabel('ACF $H_s$','Color','k')
legend('Observed',m_type)
grid on
subplot(2,1,2)
plot(lags_auto,acf_Tp_obs,'LineWidth',line_width)
hold on
plot(lags_auto,acf_Tp_sim,'LineWidth',line_width,'LineStyle','--')
%title('$T_p$ Autocorrelation')
%legend('Observed',m_type)
xlabel('Lag','Color','k')
ylabel('ACF $T_p$','Color','k')
grid on

% Cross Correlation
[xcf_sim,lags_cross] = crosscorr(Y_sim(:,1,1),Y_sim(:,2,1));
xcf_obs = crosscorr(metocean(:,3),metocean(:,4));

figure(4)
clf
plot(lags_cross,xcf_obs,'LineWidth',line_width)
hold on
plot(lags_cross,xcf_sim,'LineWidth',line_width,'LineStyle','--')
xlabel('Lag','Color','k')
ylabel('Crosscorrelation Function','Color','k')
legend('Observed',m_type)
grid on
%title('Crosscorrelation')


%% Scatter plot
series_length = length(metocean);

% Preallocation
group1 = cell(series_length,1);
group2 = cell(series_length,1);

% Create one group for model and observed
for ii = 1:series_length
    group1{ii} = 'Observed';
    group2{ii} = m_type;
end

% Vectors for scatterhist
type = cat(1,group1,group2);
HsPlot = [metocean(:,3); Y_sim(:,1,1)];
TpPlot = [metocean(:,4); Y_sim(:,2,1)];

%[0 2 4 6 8 10 12 14 16 18 20 22]
figure(5)
clf
scatterhist(TpPlot, HsPlot,'Group',type, 'MarkerSize', 2,'kernel','on',...
    'Color','br','LineWidth',[line_width,line_width],'Location','SouthWest')
xlim([0 scatter_max_x]) % 28
ylim([0 14])
xticks(0:4:scatter_max_x) % 28
%xticklabels({'2','4','6','8','10','12','14','16','18','20','22',})
yticks(0:2:14)
ylabel('$H_s$','Color','k')
xlabel('$T_p$','Color','k')
%[lgd,b] = legend('Observed',m_type);
%lgd.Location = 'none';
%set(findobj(b,'-property','MarkerSize'),'MarkerSize',10)
grid on

%% Normalized Persistence Tables

if plot_pers == 1
    
        % Total simulated time
    total_time = series_length * step_length; 

    % Size of persistence tables
    [nRow, nCol] = size(obs_above);

    % Initilize
    hindAboveTime = zeros(nRow,nCol); hindBelowTime = zeros(nRow,nCol);
    simAboveTime = zeros(nRow,nCol); simBelowTime = zeros(nRow,nCol);

    % Normalize and transform to real time
    for ii = 1:nRow
        hindAboveTime(ii,:) = 6*ii*obs_above(ii,:)/total_time;
        hindBelowTime(ii,:) = 6*ii*obs_below(ii,:)/total_time;
        simAboveTime(ii,:) = 6*ii*syn_above(ii,:)/total_time;
        simBelowTime(ii,:) = 6*ii*syn_below(ii,:)/total_time;
    end
    % Low Values Normalized Persistence ABOVE
    figure(6)
    clf
    v = [above_values(1),above_values(1)];
    [C,h] = contour(flipud(hindAboveTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on
    [C,h] = contour(flipud(simAboveTime),v,'--r','LineWidth',line_width);
    clabel(C,h)
    
    v = [above_values(2),above_values(2)];
    [C,h] = contour(flipud(hindAboveTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on 
    [C,h] = contour(flipud(simAboveTime),v,'--r','LineWidth',line_width);
    clabel(C,h)

    v = [above_values(3),above_values(3)];
    [C,h] = contour(flipud(hindAboveTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    [C,h] = contour(flipud(simAboveTime),v,'--r','LineWidth',line_width);
    clabel(C,h)

    v = [above_values(4),above_values(4)];
    [C,h] = contour(flipud(hindAboveTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on
    [C,h] = contour(flipud(simAboveTime),v,'--r','LineWidth',line_width);
    clabel(C,h)

    v = [above_values(5),above_values(5)];
    [C,h] = contour(flipud(hindAboveTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on
    [C,h] = contour(flipud(simAboveTime),v,'--r','LineWidth',line_width);
    clabel(C,h)

    v = [above_values(6),above_values(6)];
    [C,h] = contour(flipud(hindAboveTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on
    [C,h] = contour(flipud(simAboveTime),v,'--r','LineWidth',line_width);
    clabel(C,h)

    v = [above_values(7),above_values(7)];
    [C,h] = contour(flipud(hindAboveTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    [C,h] = contour(flipud(simAboveTime),v,'--r','LineWidth',line_width);
    clabel(C,h)
    yticks([1 10 20 30 40 50 60])
    yticklabels({'360','300','240','180','120','60','6'})
    xticks([1 5 9 13 17 21])
    xticklabels({'1.0','1.5','2.0','2.5','3.0','3.5'})
    xlabel('Significant Wave Height Threshold [m]','Color','k')
    ylabel('Required Window Length [h]','Color','k')
    lgd = legend({'Observed',m_type});
    lgd.Location = 'southeast';
    hold off
    grid on

    % Low Values Normalized Persistence Below
    figure(7)
    clf
    v = [below_values(1),below_values(1)];
    [C,h] = contour(flipud(hindBelowTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on
    [C,h] = contour(flipud(simBelowTime),v,'--r','LineWidth',line_width);
    clabel(C,h)
    
    v = [below_values(2),below_values(2)];
    [C,h] = contour(flipud(hindBelowTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on
    [C,h] = contour(flipud(simBelowTime),v,'--r','LineWidth',line_width);
    clabel(C,h)

    v = [below_values(3),below_values(3)];
    [C,h] = contour(flipud(hindBelowTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on
    [C,h] = contour(flipud(simBelowTime),v,'--r','LineWidth',line_width);
    clabel(C,h)
    
    v = [below_values(4),below_values(4)];
    [C,h] = contour(flipud(hindBelowTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on
    [C,h] = contour(flipud(simBelowTime),v,'--r','LineWidth',line_width);
    clabel(C,h)

    v = [below_values(5),below_values(5)];
    [C,h] = contour(flipud(hindBelowTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    hold on
    [C,h] = contour(flipud(simBelowTime),v,'--r','LineWidth',line_width);
    clabel(C,h)

    v = [below_values(6),below_values(6)];
    [C,h] = contour(flipud(hindBelowTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    [C,h] = contour(flipud(simBelowTime),v,'--r','LineWidth',line_width);
    clabel(C,h)

    v = [below_values(7),below_values(7)];
    [C,h] = contour(flipud(hindBelowTime),v,'b','LineWidth',line_width);
    clabel(C,h)
    [C,h] = contour(flipud(simBelowTime),v,'--r','LineWidth',line_width);
    clabel(C,h)
    yticks([1 10 20 30 40 50 60])
    yticklabels({'360','300','240','180','120','60','6'})
    xticks([1 5 9 13 17 21])
    xticklabels({'1.0','1.5','2.0','2.5','3.0','3.5'})
    xlabel('Significant Wave Height Threshold [m]','Color','k')
    ylabel('Required Window Length [h]','Color','k')
    lgd = legend({'Observed',m_type});
    lgd.Location = 'southeast';
    hold off
    grid on
end