import pandas as pd
import numpy as np
from sklearn.preprocessing import MinMaxScaler
from keras.models import load_model
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.stats import norm
from sklearn.metrics import mean_squared_error
import math

################################# FUNCTIONS ##################################
# convert series to supervised learning
def series_to_supervised(data, n_in=1, n_out=1, dropnan=True):
	n_vars = 1 if type(data) is list else data.shape[1]
	df = pd.DataFrame(data)
	cols, names = list(), list()
	# input sequence (t-n, ... t-1)
	for i in range(n_in, 0, -1):
		cols.append(df.shift(i))
		names += [('var%d(t-%d)' % (j+1, i)) for j in range(n_vars)]
	# forecast sequence (t, t+1, ... t+n)
	for i in range(0, n_out):
		cols.append(df.shift(-i))
		if i == 0:
			names += [('var%d(t)' % (j+1)) for j in range(n_vars)]
		else:
			names += [('var%d(t+%d)' % (j+1, i)) for j in range(n_vars)]
	# put it all together
	agg = pd.concat(cols, axis=1)
	agg.columns = names
	# drop rows with NaN values
	if dropnan:
		agg.dropna(inplace=True)
	return agg

# split supervised problem into train and test
def supervised_to_train_test(reframed,batch_size,window):
    # Transform to float
    values = reframed.values
    # Split into training and test set
    train_size = int(len(reframed) * 0.80)
    train, test = values[0:train_size], values[train_size:len(reframed)]
    # split into input and outputs
    train_X, train_Y = train[:, :-num_features], train[:, -num_features:]
    test_X, test_Y = test[:, :-num_features], test[:, -num_features:]
    # reshape input to be 3D [samples, timesteps, features]
    train_X = train_X.reshape((train_X.shape[0], window, num_features))
    test_X = test_X.reshape((test_X.shape[0], window, num_features))
    # adjust length of training and test sets to be a factor of the batch size
    num_batches_train = int(len(train_X)/batch_size)
    num_batches_test = int(len(test_X)/batch_size)    
    train_X = train_X[0:(num_batches_train*batch_size),:,:]
    train_Y = train_Y[0:(num_batches_train*batch_size),:]    
    test_X = test_X[0:(num_batches_test*batch_size),:,:]
    test_Y = test_Y[0:(num_batches_test*batch_size),:]
    # Drop months as feature
    train_Y = train_Y[:,1:]
    test_Y = test_Y[:,1:]
    return train_X, train_Y, test_X, test_Y

################################### END OF FUNC ##############################
    

# Load stationary residuals, month and year
df_observed = pd.read_csv('residuals_pred_test.csv',header=None,usecols=[0,1,2,3],
                        names=['year','month','Hs','Tp'], engine='python')


################# COPY PARAMETERS FROM BUILD SCRIPT #########################
# PARAMETERS
look_back = 32
num_features = 3

# HYPER-PARAMETERS
neurons = 128
batch_size = 64
stack_depth = 2


# Select years for training and testing
df_observed = df_observed[['month','Hs','Tp']]

# Mean and STD of observations
sigma_obs = [np.std(df_observed['Hs']), np.std(df_observed['Tp'])]
mu_obs = [np.mean(df_observed['Hs']), np.mean(df_observed['Tp'])]

# Scale all features
scaler_all = MinMaxScaler(feature_range=(0, 1))
scaled = scaler_all.fit_transform(df_observed.values)

# New scaler to only apply to Hs and Tp
scaler_pred = MinMaxScaler(feature_range=(0, 1))
scaler_pred.fit(df_observed[['Hs','Tp']].values)

# Frame as supervised learning
reframed = series_to_supervised(scaled, look_back, 1)

# Split into train and test sets and adjust length to be a factor of batch size
train_X, train_Y, test_X, test_Y = supervised_to_train_test(reframed,64,look_back)

# Load deep learning model
model = load_model('LSTM_model_1.h5')

# Re-define batch size
batch_size = 1
# --------------------- Forecast ------------------------------------------
# TEST SET

val_x = train_X[0:2920,:,:]
val_y = train_Y[0:2920,:]

# make predictions
test_predict = model.predict(val_x, batch_size=batch_size)

# Inverse scale 
test_predict = scaler_pred.inverse_transform(test_predict)
test_Y = scaler_pred.inverse_transform(val_y)

# Prediction test score
RMSE_Hs = math.sqrt(mean_squared_error(test_Y[:,0], test_predict[:,0]))
print('Test Score: %.2f RMSE' % (RMSE_Hs))

# ------------------------------ Validation -----------------------------------
sns.set_style("whitegrid")

# Prediction Hs
plt.figure(1)
plt.plot(np.arange(len(test_Y)), test_predict[:,0],'r',label="Hs Predicted")
plt.plot(np.arange(len(test_Y)), test_Y[:,0],'b',label="Hs Test")
plt.legend()
plt.show()

plt.figure(2)
plt.plot(np.arange(len(test_Y)), test_predict[:,1],'r',label="Tp Predicted")
plt.plot(np.arange(len(test_Y)), test_Y[:,1],'b',label="Tp Test")
plt.legend()
plt.show()




