import pandas as pd
import numpy as np
from sklearn.preprocessing import MinMaxScaler
from keras.models import load_model
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.stats import norm
from sklearn.metrics import mean_squared_error
import math

################################# FUNCTIONS ##################################
# convert series to supervised learning
def series_to_supervised(data, n_in=1, n_out=1, dropnan=True):
	n_vars = 1 if type(data) is list else data.shape[1]
	df = pd.DataFrame(data)
	cols, names = list(), list()
	# input sequence (t-n, ... t-1)
	for i in range(n_in, 0, -1):
		cols.append(df.shift(i))
		names += [('var%d(t-%d)' % (j+1, i)) for j in range(n_vars)]
	# forecast sequence (t, t+1, ... t+n)
	for i in range(0, n_out):
		cols.append(df.shift(-i))
		if i == 0:
			names += [('var%d(t)' % (j+1)) for j in range(n_vars)]
		else:
			names += [('var%d(t+%d)' % (j+1, i)) for j in range(n_vars)]
	# put it all together
	agg = pd.concat(cols, axis=1)
	agg.columns = names
	# drop rows with NaN values
	if dropnan:
		agg.dropna(inplace=True)
	return agg

# split supervised problem into train and test
def supervised_to_train_test(reframed,batch_size,window):
    # Transform to float
    values = reframed.values
    # Split into training and test set
    train_size = int(len(reframed) * 0.80)
    train, test = values[0:train_size], values[train_size:len(reframed)]
    # split into input and outputs
    train_X, train_Y = train[:, :-num_features], train[:, -num_features:]
    test_X, test_Y = test[:, :-num_features], test[:, -num_features:]
    # reshape input to be 3D [samples, timesteps, features]
    train_X = train_X.reshape((train_X.shape[0], window, num_features))
    test_X = test_X.reshape((test_X.shape[0], window, num_features))
    # adjust length of training and test sets to be a factor of the batch size
    num_batches_train = int(len(train_X)/batch_size)
    num_batches_test = int(len(test_X)/batch_size)    
    train_X = train_X[0:(num_batches_train*batch_size),:,:]
    train_Y = train_Y[0:(num_batches_train*batch_size),:]    
    test_X = test_X[0:(num_batches_test*batch_size),:,:]
    test_Y = test_Y[0:(num_batches_test*batch_size),:]
    # Drop months as feature
    train_Y = train_Y[:,1:]
    test_Y = test_Y[:,1:]
    return train_X, train_Y, test_X, test_Y

def simulate_LSTM(num_rep,scale_val):
    # Initilize starting window for simulation
    trainPredict = [train_X[0,:,:]]
    # Pre-allocate prediction output 
    predictions = np.zeros((num_rep,2))
    # Simulate one new prediction for the length of the simulation
    for i in range(num_rep):
        prediction = model.predict(np.array([trainPredict[-1]]), batch_size=1)
        # inverse transform to residual scale
        prediction = scaler_pred.inverse_transform(prediction)
        # Add white noise to inverse transformed predictions
        prediction[0,0] = prediction[0,0] + np.random.normal(0, scale_val[0]*sigma_obs[0], 1)
        prediction[0,1] = prediction[0,1] + np.random.normal(0, scale_val[1]*sigma_obs[1], 1)
        # Transform back to prediction scale
        prediction = scaler_pred.transform(prediction)
        # Store all predictions
        predictions[i] = prediction
        # Append last value for month to shape for next prediction
        prediction = np.insert(prediction,0,[trainPredict[-1][-1][0]])
        prediction.flatten()
        # Append predictions as the value of the previous time step
        trainPredict.append(np.vstack([trainPredict[-1][1:],prediction]))
    # Inverse transform all predictions to residual scale
    predictions = scaler_pred.inverse_transform(predictions)
    # transform to dataframe
    df_predictions = pd.DataFrame(predictions,columns =['Hs','Tp'])
    return df_predictions

################################### END OF FUNC ##############################
    

# Load stationary residuals, month and year
df_observed = pd.read_csv('residuals_rosen_M_C.csv',header=None,usecols=[0,1,2,3],
                        names=['year','month','Hs','Tp'], engine='python')


################# COPY PARAMETERS FROM BUILD SCRIPT #########################
# PARAMETERS
look_back = 16
num_features = 3

# HYPER-PARAMETERS
neurons = 128
batch_size = 64
stack_depth = 2


# Inital guess for STD scale parameters
scale_val = [1,1]

# Select years for training and testing
df_observed = df_observed[['month','Hs','Tp']]

# Mean and STD of observations
sigma_obs = [np.std(df_observed['Hs']), np.std(df_observed['Tp'])]
mu_obs = [np.mean(df_observed['Hs']), np.mean(df_observed['Tp'])]

# Scale all features
scaler_all = MinMaxScaler(feature_range=(0, 1))
scaled = scaler_all.fit_transform(df_observed.values)

# New scaler to only apply to Hs and Tp
scaler_pred = MinMaxScaler(feature_range=(0, 1))
scaler_pred.fit(df_observed[['Hs','Tp']].values)

# Frame as supervised learning
reframed = series_to_supervised(scaled, look_back, 1)

# Split into train and test sets and adjust length to be a factor of batch size
train_X, train_Y, test_X, test_Y = supervised_to_train_test(reframed,64,look_back)

# Load deep learning model
model = load_model('LSTM_model_1_H.h5')

# Re-define batch size
batch_size = 1
# --------------------- Forecast ------------------------------------------

# make predictions
test_predict = model.predict(test_X, batch_size=batch_size)

# Inverse scale 
test_predict = scaler_pred.inverse_transform(test_predict)
test_Y = scaler_pred.inverse_transform(test_Y)

# Prediction test score
RMSE_Hs = math.sqrt(mean_squared_error(test_Y[:,0], test_predict[:,0]))
print('Test Score: %.2f RMSE' % (RMSE_Hs))

# --------------------- Simulate ------------------------------------------
# Number of replications
num_rep = 37960

# Simulate
for i in range(10):
    df_predictions = simulate_LSTM(num_rep,scale_val)
    if i == 0:
        synthetic_series = df_predictions
    else:
        synthetic_series = pd.concat([synthetic_series, df_predictions],axis=1)


synthetic_series.to_csv('C:/Users/olejl/LSTM/residuals_LSTM_M_C.csv',
                  header=False, index=False)

# ------------------------------ Validation -----------------------------------
sns.set_style("whitegrid")

# Plot of series
plt.figure(1)
plt.subplot(211)
plt.plot(np.arange(num_rep),df_predictions['Hs'],'b',label="Hs Simulated")
plt.plot(np.arange(num_rep),df_observed['Hs'][:num_rep],'r',label="Hs Observed")
plt.legend()

plt.subplot(212)
plt.plot(np.arange(num_rep),df_predictions['Tp'],'b',label="Tp Simulated")
plt.plot(np.arange(num_rep),df_observed['Tp'][:num_rep],'r',
         label="Tp Observed")
plt.legend()
plt.show()

# Gaussian PDFs
# Gaussian  object for observations
obs_Hs = norm(loc = mu_obs[0], scale = sigma_obs[0])
obs_Tp = norm(loc = mu_obs[1], scale = sigma_obs[1])

# Mean and STD of predictions
sigma_pred = [np.std(df_predictions['Hs']), np.std(df_predictions['Tp'])] 
mu_pred = [np.mean(df_predictions['Hs']), np.mean(df_predictions['Tp'])] 
    
# Gaussian distribution for predictions
pred_Hs = norm(loc = mu_pred[0], scale = sigma_pred[0])
pred_Tp = norm(loc = mu_pred[1], scale = sigma_pred[1])

# Points validation of fit
x_Hs = np.arange(-10, 10, .1)
x_Tp = np.arange(-10, 10, .1)

plt.figure(2)
plt.subplot(211)
plt.plot(x_Hs, pred_Hs.pdf(x_Hs),'b',label="Hs Predicted")
plt.plot(x_Hs, obs_Hs.pdf(x_Hs),'r',label="Hs Observed")
plt.legend()

plt.subplot(212)
plt.plot(x_Tp, pred_Tp.pdf(x_Tp),'b',label="Tp Predicted")
plt.plot(x_Tp, obs_Tp.pdf(x_Tp),'r',label="Tp Observed")
plt.legend()
plt.show()

# Prediction Hs
plt.figure(3)
plt.plot(np.arange(len(test_Y)), test_predict[:,0],'b',label="Hs Predicted")
plt.plot(np.arange(len(test_Y)), test_Y[:,0],'r',label="Hs Test")
plt.legend()
plt.show()

# Scatter PLot
graph = sns.jointplot(x=df_observed.Hs, y=df_observed.Tp, color='r')
graph.x = df_predictions.Hs
graph.y = df_predictions.Tp
graph.plot_joint(plt.scatter, marker='x', c='b', s=40)



