import pandas as pd
import numpy as np
from sklearn.preprocessing import MinMaxScaler
from keras.models import Sequential
from keras.layers import Dense, LSTM, Dropout
from keras import regularizers
################################# FUNCTIONS ##################################

# Scale features 
def scale_features(metocean_data):    
    # Transform dataframe to floats
    raw_values = metocean_data.values
    #raw_values = raw_values[0:series_length,:]
    
    # Scale features
    scaler = MinMaxScaler(feature_range=(0, 1))
    scaled = scaler.fit_transform(raw_values)
    return scaler,scaled

# convert series to supervised learning
def series_to_supervised(data, n_in=1, n_out=1, dropnan=True):
	n_vars = 1 if type(data) is list else data.shape[1]
	df = pd.DataFrame(data)
	cols, names = list(), list()
	# input sequence (t-n, ... t-1)
	for i in range(n_in, 0, -1):
		cols.append(df.shift(i))
		names += [('var%d(t-%d)' % (j+1, i)) for j in range(n_vars)]
	# forecast sequence (t, t+1, ... t+n)
	for i in range(0, n_out):
		cols.append(df.shift(-i))
		if i == 0:
			names += [('var%d(t)' % (j+1)) for j in range(n_vars)]
		else:
			names += [('var%d(t+%d)' % (j+1, i)) for j in range(n_vars)]
	# put it all together
	agg = pd.concat(cols, axis=1)
	agg.columns = names
	# drop rows with NaN values
	if dropnan:
		agg.dropna(inplace=True)
	return agg

# split supervised problem into train and test
def supervised_to_train_test(reframed,batch_size,window):
    # Transform to float
    values = reframed.values
    # Split into training and test set
    train_size = int(len(reframed) * 0.80)
    train, test = values[0:train_size], values[train_size:len(reframed)]
    # split into input and outputs
    train_X, train_Y = train[:, :-num_features], train[:, -num_features:]
    test_X, test_Y = test[:, :-num_features], test[:, -num_features:]
    # reshape input to be 3D [samples, timesteps, features]
    train_X = train_X.reshape((train_X.shape[0], window, num_features))
    test_X = test_X.reshape((test_X.shape[0], window, num_features))
    # adjust length of training and test sets to be a factor of the batch size
    num_batches_train = int(len(train_X)/batch_size)
    num_batches_test = int(len(test_X)/batch_size)    
    train_X = train_X[0:(num_batches_train*batch_size),:,:]
    train_Y = train_Y[0:(num_batches_train*batch_size),:]    
    test_X = test_X[0:(num_batches_test*batch_size),:,:]
    test_Y = test_Y[0:(num_batches_test*batch_size),:]
    # Drop months as feature
    train_Y = train_Y[:,1:]
    test_Y = test_Y[:,1:]
    return train_X, train_Y, test_X, test_Y

# Network Architecture
def build_LSTM(neurons,stack_depth,batch_size):
    model = Sequential()
    if stack_depth > 1:
        for i in range(stack_depth):
            model.add(LSTM(neurons,batch_input_shape=(batch_size,
                train_X.shape[1], train_X.shape[2]),
                stateful=True,
                return_sequences=True,
                kernel_regularizer=regularizers.l2(lambda_reg)))
            model.add(Dropout(0.50))
        model.add(LSTM(neurons,batch_input_shape=(batch_size,
                    train_X.shape[1], train_X.shape[2]),
                    stateful=True,
                    kernel_regularizer=regularizers.l2(lambda_reg)))
    else:
        model.add(LSTM(neurons,batch_input_shape=(batch_size,
                    train_X.shape[1], train_X.shape[2]),
                    stateful=True,
                    kernel_regularizer=regularizers.l2(lambda_reg)))
    model.add(Dropout(0.50))
    model.add(Dense(2))
    model.compile(loss='mean_squared_error', optimizer='adam')
    return model

# Train model and return loss history
def train_LSTM(num_epochs,batch_size):
    run_history = np.zeros((num_epochs,2))
    for i in range(num_epochs):
        History = model.fit(train_X, train_Y, epochs=1, batch_size=batch_size,
                   validation_data=(test_X,test_Y), verbose=2, shuffle=False)
        model.reset_states()
        print('True Epoch:', (i+1))
        loss_dict = History.history
        run_history[i,0] = np.array(loss_dict['loss'])
        run_history[i,1] = np.array(loss_dict['val_loss'])
        run_history = np.array(run_history)
    return run_history
#################################################################

# Load the dataset
metocean_data = pd.read_csv('W:/Residuals/residuals_rosen_M_C.csv',header=None,usecols=[1,2,3],
                        names=['month','Hs','Tp'], engine='python')

# Non-variable parameters
num_features = 3
num_epochs = 60
batch_size = 64

# Scale features
scaler,scaled = scale_features(metocean_data)

# Hyper-parameter values
window = [8,16,32]
neurons = [32,64,128]
stack_depth = [1,2,3]
lambda_reg = 0.00001

# Counter for iteration number
counter = 1

for i in range(len(window)):
    for j in range(len(neurons)): 
        for k in range(len(stack_depth)):
            # Frame as supervised learning
            reframed = series_to_supervised(scaled, window[i], 1)
            
            # Split into train and test sets and adjust length to be a factor of batch size
            train_X, train_Y, test_X, test_Y = supervised_to_train_test(reframed,batch_size,window[i])
            
            # Build Network Architecture
            model = build_LSTM(neurons[j],stack_depth[k],batch_size)
            
            # Train Network 
            run_history = train_LSTM(num_epochs,batch_size)
            
            # Save loss history and model parameters 
            # Parameter column
            run_parameters = np.array([[window[i],neurons[j],stack_depth[k],lambda_reg]])
            run_parameters = np.concatenate((run_parameters,run_parameters)).T
            
            # Append loss history to parameters
            run_history = np.append(run_parameters,run_history,axis=0)
            #run_history = pd.DataFrame(run_history)
            # Add loss and parameters to complete history
            if counter == 1:
                full_history = run_history
                
            else:
                full_history = np.column_stack((full_history,run_history))
                # Write to csv after each iteration
                history_df = pd.DataFrame(full_history)
                history_df.to_csv('W:/output_pyy.csv',
                              header=False, index=False)
            counter = counter + 1
            
